#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2022  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

from waptutils import __version__

import os
import sys

try:
    wapt_root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
except:
    wapt_root_dir = 'c:/tranquilit/wapt'

import configparser
import tempfile
import platform

try:
    import distro
    linux_distribution = distro.linux_distribution(False)
except ImportError:
    linux_distribution = None

def type_windows():
    return platform.win32_ver()[0] != ''


def type_debian():
    return linux_distribution is not None and linux_distribution[0].lower() in ('debian', 'ubuntu', 'linuxmint', 'raspbian')


def type_redhat():
    return linux_distribution is not None and linux_distribution[0].lower() in ('rhel', 'centos', 'fedora', 'oracle','almalinux','rocky')


def get_distrib_major():
    if linux_distribution is not None:
        return linux_distribution[1].split('.')[0]
    else:
        raise Exception("Can't parse distrib major")


if type_debian():
    DEFAULT_WAPT_FOLDER = '/var/www/wapt'
    DEFAULT_WADS_FOLDER = '/var/www/wads'
    DEFAULT_AGENTS_FOLDER = '/var/www/wapt/waptagent'
elif type_redhat():
    DEFAULT_WAPT_FOLDER = '/var/www/html/wapt'
    DEFAULT_WADS_FOLDER = '/var/www/html/wads'
    DEFAULT_AGENTS_FOLDER = '/var/www/html/wapt/waptagent'
else:
    DEFAULT_WAPT_FOLDER = os.path.join(wapt_root_dir, 'waptserver', 'repository', 'wapt')
    DEFAULT_WADS_FOLDER = os.path.join(wapt_root_dir, 'waptserver', 'repository', 'wads')
    DEFAULT_AGENTS_FOLDER = os.path.join(wapt_root_dir, 'waptserver', 'repository', 'wapt', 'waptagent')

WAPTLOGGERS = ['waptcore', 'waptserver', 'waptserver.app', 'waptws', 'waptdb']

_defaults = {
    'client_tasks_timeout': 5,
    'clients_read_timeout': 5,
    'loglevel': 'warning',
    'secret_key': None,
    'server_uuid': None,
    'wapt_folder': DEFAULT_WAPT_FOLDER,
    'wads_folder': DEFAULT_WADS_FOLDER,
    'agents_folder': DEFAULT_AGENTS_FOLDER,
    'wapt_huey_db': None,
    'wapt_password': None,
    'wapt_user': 'admin',
    'waptserver_port': 8080,
    'waptwua_folder': '',  # default: wapt_folder + 'wua'
    'db_name': 'wapt',
    'db_host': None,
    'db_port': 5432,
    'db_user': 'wapt',
    'db_password': None,
    'db_max_connections': 90,
    'db_stale_timeout': 300,
    'db_connect_timeout': 3,
    'use_kerberos': False,
    'max_clients': 4096,
    'encrypt_host_packages': False,
    'dc_ssl_enabled': True,
    'dc_auth_enabled': False,
    'allow_unsigned_status_data': False,
    'min_password_length': 10,
    'allow_unauthenticated_registration': False,
    'allow_unauthenticated_connect': None,
    'clients_signing_key': None,
    'clients_signing_certificate': None,
    'clients_signing_crl': None,
    'clients_signing_crl_url': None,
    'clients_signing_crl_days': 30,
    'client_certificate_lifetime': 3650,
    'use_ssl_client_auth': False,
    'signature_clockskew': 5*60,
    'token_lifetime': 12*60*60,
    'application_root': '',
    'wapt_admin_group_dn': [],
    'wapt_admin_group': [],
    'ldap_auth_server': None,
    'ldap_auth_base_dn': None,
    'ldap_auth_ssl_enabled': True,
    'ldap_account_service_login': None,
    'ldap_account_service_password': None,
    'verify_cert_ldap': True,
    'htpasswd_path': None,
    'http_proxy': None,
    'nginx_http': 80,
    'nginx_https': 443,
    'remote_repo_websockets': True,
    'remote_repo_diff': False,
    'remote_repo_update_delay': 1,
    'auto_create_ldap_users': True,
    'default_ldap_users_acls': 'view',
    'default_ldap_users_perimeters': '*',
    'trusted_signers_certificates_folder': None,  # path to trusted signers certificate directory
    'trusted_users_certificates_folder': None,  # path to trusted users CA certificate directory
    'known_certificates_folder': os.path.join(os.path.dirname(DEFAULT_WAPT_FOLDER), 'ssl'),
    'wol_port': '7,9',
    'enable_store': False,
    'wapt_bind_interface': '127.0.0.1',
    'authentication_logs': True,
    'optimized_authentication_logs': True,
    'cleanup_kbs': True,
    'download_wsusscn2': True,
    'ldap_primary_group_ad_support':True,
    'ldap_nesting_group_support':True,
    'copy_winpe_x64_in_tftp_folder' : False
}

for log in WAPTLOGGERS:
    _defaults['loglevel_%s' % log] = None

DEFAULT_CONFIG_DIR = os.path.join(wapt_root_dir, 'conf')
DEFAULT_CONFIG_FILE = os.path.join(DEFAULT_CONFIG_DIR, 'waptserver.ini')


def load_config(cfgfile=DEFAULT_CONFIG_FILE):
    conf = _defaults.copy()

    # read configuration from waptserver.ini
    _config = configparser.RawConfigParser()
    if os.path.exists(cfgfile):
        _config.read(cfgfile)
    else:
        # XXX this is a kludge
        if os.getenv('USERNAME') == 'buildbot':
            return conf

    if not _config.has_section('options'):
        _config.add_section('options')

    if _config.has_option('options', 'client_tasks_timeout'):
        conf['client_tasks_timeout'] = int(_config.get('options', 'client_tasks_timeout'))

    if _config.has_option('options', 'clients_read_timeout'):
        conf['clients_read_timeout'] = int(_config.get('options', 'clients_read_timeout'))

    if _config.has_option('options', 'loglevel'):
        conf['loglevel'] = _config.get('options', 'loglevel')

    for log in WAPTLOGGERS:
        if _config.has_option('options', 'loglevel_%s' % log):
            conf['loglevel_%s' % log] = _config.get('options', 'loglevel_%s' % log)

    if _config.has_option('options', 'secret_key'):
        secret_key = _config.get('options', 'secret_key')
        if secret_key is None or len(secret_key) < 32:
            msg = 'incorrect secret_key value %s in waptserver.ini, please run postconf.py again (missing or too short)' % secret_key
            raise Exception(msg)
        conf['secret_key'] = secret_key

    if _config.has_option('options', 'server_uuid'):
        server_uuid = _config.get('options', 'server_uuid')
        if server_uuid is None or len(server_uuid) < 10:
            msg = 'incorrect server_uuid value %s in waptserver.ini, please run postconf.py again (missing or too short)' % server_uuid
            raise Exception(msg)
        conf['server_uuid'] = server_uuid

    if _config.has_option('options', 'wapt_folder'):
        conf['wapt_folder'] = _config.get('options', 'wapt_folder').rstrip('/').rstrip('\\')

    if _config.has_option('options', 'wads_folder'):
        conf['wads_folder'] = _config.get('options', 'wads_folder').rstrip('/').rstrip('\\')

    if _config.has_option('options', 'agents_folder'):
        conf['agents_folder'] = _config.get('options', 'agents_folder').rstrip('/').rstrip('\\')

    if _config.has_option('options', 'wapt_password'):
        conf['wapt_password'] = _config.get('options', 'wapt_password')

    if _config.has_option('options', 'wapt_bind_interface'):
        conf['wapt_bind_interface'] = _config.get('options', 'wapt_bind_interface')

    if _config.has_option('options', 'wapt_user'):
        conf['wapt_user'] = _config.get('options', 'wapt_user')

    if _config.has_option('options', 'waptserver_port'):
        conf['waptserver_port'] = _config.getint('options', 'waptserver_port')

    if _config.has_option('options', 'waptservice_port'):
        conf['waptservice_port'] = _config.getint('options', 'waptservice_port')

    # XXX must be processed after conf['wapt_folder']
    if _config.has_option('options', 'waptwua_folder'):
        conf['waptwua_folder'] = _config.get('options', 'waptwua_folder').rstrip('/').rstrip('\\')
    if not conf['waptwua_folder']:
        conf['waptwua_folder'] = conf['wapt_folder'] + 'wua'

    for param in ('db_name', 'db_host', 'db_user', 'db_password'):
        if _config.has_option('options', param):
            conf[param] = _config.get('options', param)

    for param in ('db_port', 'db_max_connections', 'db_stale_timeout', 'db_connect_timeout', 'max_clients'):
        if _config.has_option('options', param):
            conf[param] = _config.getint('options', param)

    if _config.has_option('options', 'use_ssl_client_auth'):
        conf['use_ssl_client_auth'] = _config.getboolean('options', 'use_ssl_client_auth')

    if _config.has_option('options', 'use_kerberos'):
        conf['use_kerberos'] = _config.getboolean('options', 'use_kerberos')

    if _config.has_option('options', 'dc_ssl_enabled'):
        conf['dc_ssl_enabled'] = _config.getboolean('options', 'dc_ssl_enabled')

    if _config.has_option('options', 'dc_auth_enabled'):
        conf['dc_auth_enabled'] = _config.getboolean('options', 'dc_auth_enabled')

    if _config.has_option('options', 'allow_unsigned_status_data'):
        conf['allow_unsigned_status_data'] = _config.getboolean('options', 'allow_unsigned_status_data')

    if _config.has_option('options', 'min_password_length'):
        conf['min_password_length'] = _config.getint('options', 'min_password_length')

    if _config.has_option('options', 'allow_unauthenticated_registration'):
        conf['allow_unauthenticated_registration'] = _config.getboolean('options', 'allow_unauthenticated_registration')

    if _config.has_option('options', 'allow_unauthenticated_connect'):
        conf['allow_unauthenticated_connect'] = _config.getboolean('options', 'allow_unauthenticated_connect')
    else:
        conf['allow_unauthenticated_connect'] = conf['allow_unauthenticated_registration']

    if _config.has_option('options', 'clients_signing_certificate'):
        conf['clients_signing_certificate'] = _config.get('options', 'clients_signing_certificate')

    if _config.has_option('options', 'clients_signing_key'):
        conf['clients_signing_key'] = _config.get('options', 'clients_signing_key')

    if _config.has_option('options', 'clients_signing_crl'):
        conf['clients_signing_crl'] = _config.get('options', 'clients_signing_crl')

    if _config.has_option('options', 'clients_signing_crl_days'):
        conf['clients_signing_crl_days'] = _config.getint('options', 'clients_signing_crl_days')

    if _config.has_option('options', 'clients_signing_crl_url'):
        conf['clients_signing_crl_url'] = _config.get('options', 'clients_signing_crl_url')

    if _config.has_option('options', 'client_certificate_lifetime'):
        conf['client_certificate_lifetime'] = _config.getint('options', 'client_certificate_lifetime')

    if _config.has_option('options', 'signature_clockskew'):
        conf['signature_clockskew'] = _config.getint('options', 'signature_clockskew')

    if _config.has_option('options', 'token_lifetime'):
        conf['token_lifetime'] = _config.getint('options', 'token_lifetime')

    if _config.has_option('options', 'application_root'):
        conf['application_root'] = _config.get('options', 'application_root')

    if _config.has_option('options', 'wapt_huey_db'):
        conf['wapt_huey_db'] = _config.get('options', 'wapt_huey_db')
    elif conf['application_root']:
        conf['wapt_huey_db'] = os.path.join(wapt_root_dir, 'db', '%s.sqlite' % conf['application_root'])
    else:
        conf['wapt_huey_db'] = os.path.join(wapt_root_dir, 'db', 'waptservertasks.sqlite')

    for option in ('wapt_admin_group_dn', 'ldap_auth_server', 'ldap_auth_base_dn'):
        if _config.has_option('options', option):
            conf[option] = _config.get('options', option)

    if _config.has_option('options', 'ldap_account_service_login'):
        conf['ldap_account_service_login'] = _config.get('options', 'ldap_account_service_login')

    if _config.has_option('options', 'ldap_account_service_password'):
        conf['ldap_account_service_password'] = _config.get('options', 'ldap_account_service_password')

    if _config.has_option('options', 'wapt_admin_group'):
        conf['wapt_admin_group'] = [elem.strip() for elem in _config.get('options', 'wapt_admin_group').split(',')]

    if _config.has_option('options', 'ldap_auth_ssl_enabled'):
        conf['ldap_auth_ssl_enabled'] = _config.getboolean('options', 'ldap_auth_ssl_enabled')

    if _config.has_option('options', 'ldap_primary_group_ad_support'):
        conf['ldap_primary_group_ad_support'] = _config.getboolean('options', 'ldap_primary_group_ad_support')

    if _config.has_option('options', 'ldap_nesting_group_support'):
        conf['ldap_nesting_group_support'] = _config.getboolean('options', 'ldap_nesting_group_support')

    if _config.has_option('options', 'copy_winpe_x64_in_tftp_folder'):
        conf['copy_winpe_x64_in_tftp_folder'] = _config.getboolean('options', 'copy_winpe_x64_in_tftp_folder')

    if _config.has_option('options', 'verify_cert_ldap'):
        conf['verify_cert_ldap'] = _config.getboolean('options', 'verify_cert_ldap')

    if _config.has_option('options', 'default_ldap_users_acls'):
        conf['default_ldap_users_acls'] = _config.get('options', 'default_ldap_users_acls')

    if _config.has_option('options', 'default_ldap_users_perimeters_sha1'):
        conf['default_ldap_users_perimeters_sha1'] = _config.get('options', 'default_ldap_users_perimeters_sha1')

    if _config.has_option('options', 'http_proxy'):
        conf['http_proxy'] = _config.get('options', 'http_proxy')

    # path to a htpasswd file for additional users authentication (register/unregister)
    if _config.has_option('options', 'htpasswd_path'):
        conf['htpasswd_path'] = _config.get('options', 'htpasswd_path')

    if _config.has_option('options', 'remote_repo_update_delay'):
        conf['remote_repo_update_delay'] = _config.get('options', 'remote_repo_update_delay')

    if _config.has_option('options', 'remote_repo_diff'):
        conf['remote_repo_diff'] = _config.getboolean('options', 'remote_repo_diff')

    if _config.has_option('options', 'remote_repo_websockets'):
        conf['remote_repo_websockets'] = _config.getboolean('options', 'remote_repo_websockets')

    # path to X509 certificates file of trusted signers to restrict access to upload packages / actions proxying
    if _config.has_option('options', 'trusted_signers_certificates_folder'):
        conf['trusted_signers_certificates_folder'] = _config.get('options', 'trusted_signers_certificates_folder')

    # path to collected known certificates and associated CRL
    if _config.has_option('options', 'known_certificates_folder'):
        conf['known_certificates_folder'] = _config.get('options', 'known_certificates_folder')
    else:
        conf['known_certificates_folder'] = os.path.join(os.path.dirname(conf['wapt_folder']), 'ssl')

    if _config.has_option('options', 'wol_port'):
        conf['wol_port'] = _config.get('options', 'wol_port')

    if _config.has_option('options', 'enable_store'):
        conf['enable_store'] = _config.getboolean('options', 'enable_store')

    if _config.has_option('options', 'authentication_logs'):
        conf['authentication_logs'] = _config.getboolean('options', 'authentication_logs')

    if _config.has_option('options', 'optimized_authentication_logs'):
        conf['optimized_authentication_logs'] = _config.getboolean('options', 'optimized_authentication_logs')

    if _config.has_option('options' , 'cleanup_kbs'):
        conf['cleanup_kbs'] = _config.getboolean('options', 'cleanup_kbs')

    if _config.has_option('options' , 'download_wsusscn2'):
        conf['download_wsusscn2'] = _config.getboolean('options', 'download_wsusscn2')

    return conf


def write_config_file(cfgfile=DEFAULT_CONFIG_FILE, server_config=None, non_default_values_only=True):
    if server_config is None:
        server_config = _defaults.copy()

    # read configuration from waptserver.ini
    _config = configparser.RawConfigParser()
    if os.path.isfile(cfgfile):
        _config.read(cfgfile)
        actual_server_config = load_config(cfgfile)
    else:
        actual_server_config = _defaults

    if not _config.has_section('options'):
        _config.add_section('options')

    for key in server_config:
        if (not non_default_values_only) or server_config[key] != actual_server_config.get(key,_defaults.get(key)) :
            if server_config[key] is None:
                _config.remove_option('options', key)
            else:
                if isinstance(server_config[key], list):
                    _config.set('options', key, ','.join(server_config[key]))
                else:
                    _config.set('options', key, server_config[key])

    if not os.path.isdir(os.path.dirname(cfgfile)):
        os.makedirs(os.path.dirname(cfgfile))

    with open(cfgfile, 'w') as inifile:
        _config.write(inifile)


def rewrite_config_item(cfg_file=DEFAULT_CONFIG_FILE, *args):
    config = configparser.RawConfigParser()

    with open(cfg_file, 'r') as cfg:
        cfg_backup = cfg.read()

    try:
        config.read(cfg_file)
        config.set(*args)
        with open(cfg_file, 'w') as cfg:
            config.write(cfg)
    except Exception as e:
        print('Couldn\'t rewrite config item {} : {}'.format(args, e), file=sys.stderr)
        with open(cfg_file, 'w') as cfg:
            cfg.write(cfg_backup)

def get_http_proxies(config):
    if config.get('http_proxy') is not None:
        http_proxies = {'http': config['http_proxy'], 'https': config['http_proxy']}
    else:
        http_proxies = None
    return http_proxies
