#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2022  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##
import os
import sys

from waptserver.config import __version__

import email.utils

from flask import request, send_from_directory, Blueprint, session

import hashlib
import json
import random
import re
import requests
import shutil
import stat
import subprocess
import time
import traceback
import uuid
import logging
import zlib

import waptserver.config

from waptserver.utils import *
from waptserver.model import *

from waptserver.common import make_response, make_response_from_exception
from waptserver.decorators import requires_auth, wapt_db_readonly, require_wapt_db, requires_enterprise

try:
    # i18n
    from flask_babel import Babel
    from flask_babel import gettext
except ImportError:
    gettext = (lambda s: s)
    Babel = None

_ = gettext

enterprise = Blueprint('enterprise', __name__)


def prepare_request_data(request):
    if request.headers.get("Content-Encoding") != "gzip":
        return
    data = zlib.decompress(request.data)
    request.data = data


@enterprise.route('/api/v3/normalization_import_softwares', methods=['GET'])
@requires_auth(['admin', 'view'])
@requires_enterprise
@require_wapt_db
def normalization_import_softwares():
    sql = """
    INSERT INTO normalization ( original_name, key, banned, windows_update, created_on, updated_on )
    SELECT
	case WHEN (S.name is null) or (S.name = '') THEN S.key ELSE S.Name END as original_name,
	S.key               as key,
	false               as banned,
	false               as windows_update,
        current_timestamp   as created_on,
        current_timestamp   as updated_on
    FROM
	normalization N
	RIGHT join ( SELECT name, key  FROM hostsoftwares GROUP BY name, key ) S ON (N.original_name = S.name) AND (N.key = S.KEY)
    WHERE
	N.original_name is NULL;
    """
    start_time = time.time()
    with wapt_db.atomic() as trans:
        try:
            c = wapt_db.execute_sql(sql)
            r = make_response(result=c.rowcount,  success=True, request_time=time.time() - start_time)
            return r
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@enterprise.route('/api/v3/normalization_filter', methods=['POST'])
@requires_auth(['admin', 'view'])
@requires_enterprise
@require_wapt_db
def normalization_filter():
    start_time = time.time()
    prepare_request_data(request)
    data = ujson.loads(request.data)

    limit = int(data.get("limit", 1000))

    try:
        filter_string = data.get("filter_string", "").lower()
        logger.debug("filter string = " + filter_string)
        if data.get("only_kbs", False):
            data = Normalization.select(Normalization.key, Normalization.normalized_name, Normalization.banned, Normalization.windows_update,
            SQL("""CASE WHEN (original_name is null) or (original_name = '') THEN key ELSE original_name END""").alias('original_name')).where(
            Normalization.windows_update == True)
        else:
            data = Normalization.select(Normalization.key, Normalization.normalized_name, Normalization.banned, Normalization.windows_update,
            SQL("""CASE WHEN (original_name is null) or (original_name = '') THEN key ELSE original_name END""").alias('original_name'))

        if len(filter_string) > 0:
            data = data.where(
                fn.LOWER(Normalization.original_name).contains(filter_string) |
                fn.LOWER(Normalization.key).contains(filter_string) |
                fn.LOWER(Normalization.normalized_name).contains(filter_string)
            )

        data = data.order_by(Normalization.original_name)
        data = data.limit(limit)
        data = list(data.dicts())

        r = make_response({'softwares' : data, 'softwares_count' : len(data)} , success=True, request_time=start_time - time.time())
        return r
    except Exception as e:
        return make_response_from_exception(e)


@enterprise.route('/api/v3/normalization_write', methods=['POST'])
@requires_auth(['admin', 'run_report'])
@requires_enterprise
@require_wapt_db
def normalization_write():
    start_time = time.time()
    prepare_request_data(request)
    data = ujson.loads(request.data)
    with wapt_db.atomic() as trans:
        try:
            for l in data:
                n = Normalization()
                n.original_name = l['original_name']
                n.key = l['key']
                n.normalized_name = l['normalized_name']
                n.banned = l['banned']
                n.windows_update = l['windows_update']
                n.save()
            trans.commit()
            data = [len(data)]
            r = make_response(data, msg='normalization_write', success=True, request_time=start_time - time.time())
            return r
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@enterprise.route('/api/v3/reporting_exec', methods=['GET'])
@requires_auth(['admin', 'run_report'])
@requires_enterprise
@wapt_db_readonly
def reporting_exec():
    start_time = time.time()
    R = None
    with wapt_db.atomic() as trans:
        try:
            id = request.args.get("id", None)
            if id is None:
                raise Exception("Param id is recquired")

            query = ReportingQueries.select(ReportingQueries.query).where(ReportingQueries.id == id)
            query = list(query.dicts())
            if 0 == len(query):
                raise Exception("Query id %d not found" % (id))
            data = ReportingQueries.raw(query[0]["query"])
            data = list(data.dicts())
            R = make_response(data, success=True, request_time=time.time() - start_time)
        except Exception as e:
            R = make_response_from_exception(e)
    return R


@enterprise.route('/api/v3/reporting_exec_raw', methods=['GET', 'POST'])
@requires_auth(['admin', 'run_report'])
@requires_enterprise
@wapt_db_readonly
def reporting_exec_raw():
    start_time = time.time()
    prepare_request_data(request)
    R = None
    with wapt_db.atomic() as trans:
        try:
            query = ujson.loads(request.data)
            if not query:
                raise Exception("Param query is recquired")

            sql = query['query']
            if not sql:
                raise Exception("Param query is recquired")

            data = ReportingQueries.raw(sql)
            data = list(data.dicts())
            R = make_response(data, success=True, request_time=time.time() - start_time)
        except Exception as e:
            trans.rollback()
            R = make_response_from_exception(e)
    return R


@enterprise.route('/api/v3/reporting_list', methods=['GET'])
@requires_auth(['admin', 'run_report'])
@requires_enterprise
@wapt_db_readonly
def reporting_list():
    start_time = time.time()
    wapt_db_connect()
    try:
        data = ReportingQueries.select(ReportingQueries.id, ReportingQueries.name, ReportingQueries.query, ReportingQueries.settings,
            ReportingQueries.as_host_uuid, ReportingQueries.as_package_uuid, ReportingQueries.as_kb_update_id, ReportingQueries.as_software_key).order_by(ReportingQueries.name)
        data = list(data.dicts())
        return make_response(data, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)


@enterprise.route('/api/v3/reporting_save', methods=['GET', 'POST'])
@requires_auth(['admin', 'edit_report'])
@requires_enterprise
@require_wapt_db
def reporting_save():
    start_time = time.time()
    prepare_request_data(request)
    try:
        queries_list = ujson.loads(request.data)
        if not queries_list:
            raise Exception("reporting : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as trans:
        try:
            data = []
            for query in queries_list:
                rq = ReportingQueries()
                if "id" in query:
                    if query["id"] >= 0:
                        rq.id = query["id"]
                if "name" in query:
                    rq.name = query["name"]
                if "query" in query:
                    rq.query = query["query"]
                if "settings" in query:
                    rq.settings = query["settings"]
                if "as_host_uuid" in query:
                    rq.as_host_uuid = query["as_host_uuid"]
                if "as_package_uuid" in query:
                    rq.as_package_uuid = query["as_package_uuid"]
                if "as_kb_update_id" in query:
                    rq.as_kb_update_id = query["as_kb_update_id"]
                if "as_software_key" in query:
                    rq.as_software_key = query["as_software_key"]

                rq.save()
                data.append({query["id"]: rq.id})
            trans.commit()
            R = make_response(data, success=True, request_time=time.time() - start_time)
            return R
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@enterprise.route('/api/v3/reporting_delete', methods=['GET', 'POST'])
@requires_auth(['admin', 'edit_report'])
@requires_enterprise
@require_wapt_db
def reporting_delete():
    start_time = time.time()
    prepare_request_data(request)
    try:
        id_list = ujson.loads(request.data)
        if not id_list:
            raise Exception("reporting : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as trans:
        try:
            data = []
            cnt = ReportingQueries.delete().where(ReportingQueries.id.in_(id_list)).execute()
            trans.commit()
            R = make_response(cnt, success=True, request_time=time.time() - start_time)
            return R
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


# Users

@enterprise.route('/api/v3/waptusers', methods=['GET', 'PUT'])
@requires_auth(['admin'])
@requires_enterprise
@wapt_db_readonly
def waptusers_get():
    start_time = time.time()
    if request.method == 'GET':
        try:
            data = list(WaptUsers.select().dicts())
            R = make_response(data, success=True, request_time=time.time() - start_time, msg='%s Wapt users' % len(data))
            return R
        except Exception as e:
            return make_response_from_exception(e)
    elif request.method == 'PUT':
        # we have a single delta to apply
        prepare_request_data(request)
        try:
            delta = ujson.loads(request.data)
            if not delta:
                raise Exception("No delta rows list supplied")
        except Exception as e:
            return make_response_from_exception(e)

        with wapt_db.atomic() as trans:
            try:
                result = []
                for row in delta:
                    id = row.get('id')
                    if id is not None and id >= 0:
                        # get existing
                        user_rec = WaptUsers.get(id=id)
                    else:
                        # create new
                        user_rec = WaptUsers()
                    for (k, v) in row.items():
                        if k != 'id' and hasattr(WaptUsers, k):
                            setattr(user_rec, k, v)
                    user_rec.save()
                    result.append(user_rec)
                trans.commit()
                R = make_response(result, success=True, request_time=time.time() - start_time)
                return R
            except Exception as e:
                trans.rollback()
                return make_response_from_exception(e)


@enterprise.route('/api/v3/waptusers', methods=['POST'])
@requires_auth(['admin'])
@requires_enterprise
@require_wapt_db
def waptusers_post():
    DataClass = WaptUsers
    start_time = time.time()
    prepare_request_data(request)
    try:
        changes = ujson.loads(request.data)
        if not changes:
            raise Exception("No changes data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as trans:
        try:
            result = []
            for change in changes:
                update_type = change.get('update_type')
                id = change.get('id')
                old_values = change.get('old_values')
                new_values = change.get('new_values')

                # insert
                if update_type in (1, 2):
                    if id is not None and id >= 0:
                        # get existing
                        rec = DataClass.get(id=id)
                    else:
                        # create new
                        rec = DataClass(**new_values)
                    # init with new values
                    if new_values:
                        for (k, v) in new_values.items():
                            if k != 'id' and hasattr(DataClass, k):
                                setattr(rec, k, v)
                    new_id = rec.save()
                    result.append((id, new_id))
                # delete
                elif update_type == 3:
                    DataClass.delete_by_id(id)
                    result.append((id, None))

            trans.commit()
            R = make_response(result, success=True, request_time=time.time() - start_time)
            return R
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@enterprise.route('/api/v3/waptusers_delete', methods=['POST'])
@requires_auth(['admin'])
@requires_enterprise
@require_wapt_db
def waptusers_delete():
    start_time = time.time()
    prepare_request_data(request)
    try:
        id_list = ujson.loads(request.data)
        if not id_list:
            raise Exception("No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as trans:
        try:
            data = []
            cnt = WaptUsers.delete().where(WaptUsers.id.in_(id_list)).execute()
            trans.commit()
            return make_response(cnt, success=True, request_time=time.time() - start_time, msg='%s users deleted' % len(id_list))
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@enterprise.route('/api/v3/host_plugins', methods=['GET'])
@enterprise.route('/api/v3/import_plugins', methods=['GET']) # back compat
@requires_auth(['admin'])
@requires_enterprise
@require_wapt_db
def hosts_plugins():
    wapt_db_connect()
    try:
        data = list(Plugins.select().dicts())
        return make_response(data, success=True)
    except Exception as e:
        return make_response_from_exception(e)

@enterprise.route('/api/v3/save_plugins', methods=['POST'])
@requires_auth(['admin'])
@requires_enterprise
@require_wapt_db
def save_plugins():
    start_time = time.time()
    prepare_request_data(request)
    try:
        plugin_list = ujson.loads(request.data)
        if not plugin_list:
            raise Exception("Plugins : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as db:
        try:
            data = []
            for plugin in plugin_list:
                # updates automatically the existing data ??
                pl = Plugins(**{k:plugin[k] for k in plugin if hasattr(Plugins,k)})
                pl.save()
                data.append(dict(pl))
            db.commit()
            return make_response(data, success=True, request_time=time.time() - start_time)

        except Exception as e:
            db.rollback()
            return make_response_from_exception(e)

@enterprise.route('/api/v3/delete_plugin', methods=['GET','POST'])
@requires_auth(['admin'])
@requires_enterprise
@require_wapt_db
def delete_plugin():
    start_time = time.time()
    prepare_request_data(request)
    try:
        plugin_to_delete = ujson.loads(request.data)
        if not plugin_to_delete:
            raise Exception("Plugins : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as db:
        try:
            cnt = Plugins.delete().where(Plugins.name.in_(plugin_to_delete)).execute()
            return make_response(cnt, success=True, request_time=time.time() - start_time)
        except Exception as e:
            db.rollback()
            return make_response_from_exception(e)

@enterprise.route('/api/v3/delete_deploy_config', methods=['GET','POST'])
@requires_auth(['admin', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def delete_deploy_config():
    start_time = time.time()
    prepare_request_data(request)
    try:
        config_to_delete = ujson.loads(request.data)
        if not config_to_delete:
            raise Exception("Deploy Config : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as db:
        try:
            cnt = DeployConfig.delete().where(DeployConfig.id.in_(config_to_delete)).execute()
            return make_response(cnt, success=True, request_time=time.time() - start_time)
        except Exception as e:
            db.rollback()
            return make_response_from_exception(e)

@enterprise.route('/api/v3/get_deploy_config', methods=['GET'])
@requires_auth(['admin', 'wads_view', 'wads_host_deploy', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def get_deploy_config():
    start_time = time.time()
    wapt_db_connect()
    try:
        data = DeployConfig.select()
        data = list(data.dicts())
        return make_response(data, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)

@enterprise.route('/api/v3/save_deploy_config', methods=['POST'])
@requires_auth(['admin', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def save_deploy_config():
    start_time = time.time()

    prepare_request_data(request)
    delta_recs = ujson.loads(request.data)

    try:
        result = update_or_insert(DeployConfig, delta_recs,session.get('user'))
        return make_response(result=result, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)


@enterprise.route('/api/v3/delete_host_deploy', methods=['GET','POST'])
@requires_auth(['admin', 'wads_host_deploy', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def delete_host_deploy():
    start_time = time.time()
    prepare_request_data(request)
    try:
        host_to_delete = ujson.loads(request.data)
        if not host_to_delete:
            raise Exception("Host to Deploy : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as db:
        try:
            cnt = HostOSDeploy.delete().where(HostOSDeploy.id.in_(host_to_delete)).execute()
            return make_response(cnt, success=True, request_time=time.time() - start_time)
        except Exception as e:
            db.rollback()
            return make_response_from_exception(e)

@enterprise.route('/api/v3/import_host_from_inventory', methods=['GET'])
@requires_auth(['admin', 'wads_host_deploy', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def import_host_from_inventory():
    start_time = time.time()
    wapt_db_connect()
    try:
        list_mac = []
        data = HostOSDeploy.select(HostOSDeploy.mac_addresses).dicts()
        for entry in data:
            list_mac.extend([mac.lower() for mac in entry['mac_addresses'].split(',')])

        data = list(Hosts.select().dicts())
        dict_hostname_mac = {}
        for entry in data:
            try:
                doublon = False
                new_mac_list = []
                if entry['wmi'] :
                    for mac in entry['wmi']['Win32_NetworkAdapter'] :
                        if 'wifi' in str(mac).lower() or "bluetooth" in str(mac).lower() or 'wireless' in str(mac).lower() or 'wi-fi' in str(mac).lower() or 'sans fil' in str(mac).lower() :
                            continue
                        if not mac.get('PhysicalAdapter'):
                            continue
                        if mac['PNPDeviceID'].startswith('ROOT\\'):
                            continue
                        maclower = mac['MACAddress'].lower()
                        if not maclower in list_mac:
                            new_mac_list.append(maclower)
                            list_mac.append(maclower)
                        else:
                            doublon =True
                            break
                else:
                    for mac in entry['mac_addresses']:
                        maclower = mac.lower()
                        if not maclower in list_mac:
                            new_mac_list.append(maclower)
                            list_mac.append(maclower)
                        else:
                            doublon =True
                            break
                if doublon:
                    continue
                if not new_mac_list:
                    continue
                dict_hostname_mac[entry['computer_name']]=new_mac_list
                serialnr = ''
                manufacturer= ''
                product = ''
                uuid = ''
                try:
                    uuid = entry['dmi']['System_Information']['UUID']
                    if not 'Base_Board_Information' in entry['dmi']:
                         serialnr = entry['serialnr']
                         manufacturer= entry['manufacturer']
                         product = entry['host_info']['system_productname']
                    else:
                         serialnr = entry['dmi']['Base_Board_Information']['Serial_Number']
                         manufacturer = entry['dmi']['Base_Board_Information']['Manufacturer']
                         product = entry['dmi']['Base_Board_Information']['Product_Name']
                except:
                    pass

                host_data = dict(
                        mac_addresses=','.join(new_mac_list),
                        hostname=entry['computer_name'].split('.')[0],
                        uuid=uuid,
                        manufacturer=manufacturer,
                        serial_number=serialnr,
                        waiting_to_deploy=False,
                        djoin="",
                        status="imported from inventory",
                        product = product
                )

                HostOSDeploy.insert(**host_data).execute()
            except:
                pass
        data  = dict_hostname_mac
        return make_response(data, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)

@enterprise.route('/api/v3/import_host_from_json', methods=['POST'])
@requires_auth(['admin', 'wads_host_deploy', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def import_host_from_json():
    start_time = time.time()
    prepare_request_data(request)
    try:
        hosts_received = ujson.loads(request.data)
        if not hosts_received:
            raise Exception("Hosts : No data to import")
    except Exception as e:
        return make_response_from_exception(e)
    wapt_db_connect()
    try:
        existing_macs = []
        data = list(HostOSDeploy.select(HostOSDeploy.mac_addresses).tuples())
        for entry in data:
            existing_macs.extend([mac.lower() for mac in entry[0].split(',')])
        new_hosts = []
        for new_host in hosts_received:
            if len(new_host['hostname']) > 15:
                continue
            try:
                new_host['mac_addresses'] = new_host['mac_addresses'].replace('-',':').lower()
                new_host_macs = new_host['mac_addresses'].split(',')
                for new_host_mac in new_host_macs:
                    new_host_mac = new_host_mac.lower()
                    if not new_host_mac in existing_macs:
                        make_insert = True
                        if len(new_host_macs) > 1:
                            for the_mac in new_host_macs:
                                the_mac = the_mac.lower()
                                if the_mac in existing_macs:
                                    make_insert = False
                                    break
                        if make_insert:
                            new_host['mac_addresses'] = new_host['mac_addresses'].lower()
                            HostOSDeploy.insert(**new_host).execute()
                            existing_macs.extend(new_host_macs)
                            new_hosts.append(new_host)
                            break

            except Exception as e:
                print('erreur: %s' % e)

        data  = new_hosts
        return make_response(data, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)

@enterprise.route('/api/v3/get_host_deploy', methods=['GET'])
@requires_auth(['admin', 'wads_admin', 'wads_host_deploy', 'wads_view'])
@requires_enterprise
@require_wapt_db
def get_host_deploy():
    start_time = time.time()
    wapt_db_connect()
    try:
        data = list(HostOSDeploy.select().dicts())
        return make_response(data, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)


def update_or_insert(Model,delta_recs,user):
    with wapt_db.atomic() as db:
        result = {}
        for rec_data in delta_recs:
            # id column is managed by database, fields starting with '_' are calculated fields.
            data = {k: rec_data[k] for k in rec_data if k in Model._meta.columns and k != 'id' and not k.startswith('_')}
            # negative id are client side generated temporary IDs for new inserted records.
            if rec_data.get('id',-1) < 0:
                data['created_on'] = datetime.datetime.now()
                data['created_by'] = user
                created = Model.insert(**data).returning(Model.id).execute()
                if rec_data.get('id'):
                    result[rec_data.get('id')] = created[0].id
            else:
                # data comes from an existing record.
                data['updated_on'] = datetime.datetime.now()
                data['updated_by'] = user
                update_count = Model.update(**data).where(Model.id == rec_data['id']).execute()
                if update_count > 0:
                    result[rec_data.get('id')] = rec_data.get('id')
        db.commit()
    return result

@enterprise.route('/api/v3/save_host_deploy', methods=['POST'])
@requires_auth(['admin', 'wads_host_deploy', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def save_host_deploy():
    start_time = time.time()

    prepare_request_data(request)
    hosts = ujson.loads(request.data)

    try:
        result = update_or_insert(HostOSDeploy, hosts, session.get('user'))
        return make_response(result=result, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)

@enterprise.route('/api/v3/delete_iso_config', methods=['GET','POST'])
@requires_auth(['admin'])
@requires_enterprise
@require_wapt_db
def delete_iso_config():
    start_time = time.time()
    prepare_request_data(request)
    try:
        iso_to_delete = ujson.loads(request.data)
        if not iso_to_delete:
            raise Exception("Iso : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as db:
        try:
            cnt = IsoConfig.delete().where(IsoConfig.id.in_(iso_to_delete)).execute()
            return make_response(cnt, success=True, request_time=time.time() - start_time)
        except Exception as e:
            db.rollback()
            return make_response_from_exception(e)

@enterprise.route('/api/v3/get_iso_config', methods=['GET'])
@requires_auth(['admin', 'wads_view', 'wads_host_deploy', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def get_iso_config():
    start_time = time.time()
    wapt_db_connect()
    try:
        data = IsoConfig.select()
        data = list(data.dicts())
        return make_response(data, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)


@enterprise.route('/api/v3/save_iso_config', methods=['POST'])
@requires_auth(['admin', 'wads_admin'])
@requires_enterprise
@require_wapt_db
def save_iso_config():
    start_time = time.time()

    prepare_request_data(request)
    delta_recs = ujson.loads(request.data)

    try:
        result = update_or_insert(IsoConfig, delta_recs, session.get('user'))
        return make_response(result=result, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)
